@@ -49,6 +49,7 @@
 
 /* Underlined */
 #define ANSI_GREY_UNDERLINE              "\x1B[0;4;38;5;245m"
+#define ANSI_BRIGHT_BLACK_UNDERLINE      "\x1B[0;4;90m"
 #define ANSI_HIGHLIGHT_RED_UNDERLINE     "\x1B[0;1;4;31m"
 #define ANSI_HIGHLIGHT_GREEN_UNDERLINE   "\x1B[0;1;4;32m"
 #define ANSI_HIGHLIGHT_YELLOW_UNDERLINE  "\x1B[0;1;4;38;5;185m"
@@ -62,8 +63,13 @@
 #define ANSI_HIGHLIGHT_UNDERLINE "\x1B[0;1;4m"
 
 /* Fallback colors: 256 -> 16 */
-#define ANSI_HIGHLIGHT_GREY_FALLBACK   "\x1B[0;1;90m"
-#define ANSI_HIGHLIGHT_YELLOW_FALLBACK "\x1B[0;1;33m"
+#define ANSI_HIGHLIGHT_GREY_FALLBACK             "\x1B[0;1;90m"
+#define ANSI_HIGHLIGHT_GREY_FALLBACK_UNDERLINE   "\x1B[0;1;4;90m"
+#define ANSI_HIGHLIGHT_YELLOW_FALLBACK           "\x1B[0;1;33m"
+#define ANSI_HIGHLIGHT_YELLOW_FALLBACK_UNDERLINE "\x1B[0;1;4;33m"
+
+/* Background colors */
+#define ANSI_BACKGROUND_BLUE "\x1B[44m"
 
 /* Reset/clear ANSI styles */
 #define ANSI_NORMAL "\x1B[0m"
@@ -80,6 +86,7 @@
 #if 0 /// UNNEEDED by elogind
 int reset_terminal_fd(int fd, bool switch_to_text);
 int reset_terminal(const char *name);
+int set_terminal_cursor_position(int fd, unsigned int row, unsigned int column);
 #endif // 0
 
 int open_terminal(const char *name, int mode);
@@ -113,7 +120,7 @@ typedef enum ColorMode {
         /* Only 256 colors. */
         COLOR_256,
 
-        /* For truecolor or 24bit color support.*/
+        /* For truecolor or 24bit color support. */
         COLOR_24BIT,
 
         _COLOR_INVALID = -EINVAL,
@@ -128,6 +135,7 @@ int terminal_vhangup(const char *name);
 
 int terminal_set_size_fd(int fd, const char *ident, unsigned rows, unsigned cols);
 #endif // 0
+int proc_cmdline_tty_size(const char *tty, unsigned *ret_rows, unsigned *ret_cols);
 
 int chvt(int vt);
 
@@ -193,21 +201,25 @@ static inline bool colors_enabled(void) {
                 }                                              \
         }
 
-#define DEFINE_ANSI_FUNC_UNDERLINE(name, NAME)                            \
-        static inline const char *ansi_##name(void) {                     \
-                return underline_enabled() ? ANSI_##NAME ANSI_UNDERLINE : \
-                        colors_enabled() ? ANSI_##NAME : "";              \
+static inline const char *ansi_underline(void) {
+        return underline_enabled() ? ANSI_UNDERLINE : ANSI_NORMAL;
+}
+
+#define DEFINE_ANSI_FUNC_UNDERLINE(name, NAME)                          \
+        static inline const char *ansi_##name(void) {                   \
+                return underline_enabled() ? ANSI_##NAME##_UNDERLINE :  \
+                        colors_enabled() ? ANSI_##NAME : "";            \
         }
 
 
 #if 0 /// UNNEEDED by elogind
-#define DEFINE_ANSI_FUNC_UNDERLINE_256(name, NAME, FALLBACK)                                                           \
-        static inline const char *ansi_##name(void) {                                                                  \
-                switch (get_color_mode()) {                                                                            \
-                        case COLOR_OFF: return "";                                                                     \
-                        case COLOR_16: return underline_enabled() ? ANSI_##FALLBACK ANSI_UNDERLINE : ANSI_##FALLBACK;  \
-                        default : return underline_enabled() ? ANSI_##NAME ANSI_UNDERLINE: ANSI_##NAME;                \
-                }                                                                                                      \
+#define DEFINE_ANSI_FUNC_UNDERLINE_256(name, NAME, FALLBACK)                                                        \
+        static inline const char *ansi_##name(void) {                                                               \
+                switch (get_color_mode()) {                                                                         \
+                        case COLOR_OFF: return "";                                                                  \
+                        case COLOR_16: return underline_enabled() ? ANSI_##FALLBACK##_UNDERLINE : ANSI_##FALLBACK;  \
+                        default : return underline_enabled() ? ANSI_##NAME##_UNDERLINE: ANSI_##NAME;                \
+                }                                                                                                   \
         }
 #endif // 0
 
@@ -232,7 +244,9 @@ DEFINE_ANSI_FUNC(bright_black,      BRIGHT_BLACK);
 DEFINE_ANSI_FUNC(bright_red,        BRIGHT_RED);
 DEFINE_ANSI_FUNC(bright_green,      BRIGHT_GREEN);
 DEFINE_ANSI_FUNC(bright_yellow,     BRIGHT_YELLOW);
+#endif // 0
 DEFINE_ANSI_FUNC(bright_blue,       BRIGHT_BLUE);
+#if 0 /// UNNEEDED by elogind
 DEFINE_ANSI_FUNC(bright_magenta,    BRIGHT_MAGENTA);
 DEFINE_ANSI_FUNC(bright_cyan,       BRIGHT_CYAN);
 DEFINE_ANSI_FUNC(bright_white,      BRIGHT_WHITE);
@@ -255,7 +269,6 @@ static inline const char* _ansi_highlight_yellow(void) {
 }
 #endif // 0
 
-DEFINE_ANSI_FUNC_UNDERLINE(underline,                       NORMAL);
 #if 0 /// UNNEEDED by elogind
 DEFINE_ANSI_FUNC_UNDERLINE(highlight_underline,             HIGHLIGHT);
 DEFINE_ANSI_FUNC_UNDERLINE_256(grey_underline,              GREY, BRIGHT_BLACK);
