@@ -75,9 +75,45 @@ static int elogind_signal_handler(
 }
 
 
+/* This handler is caught on SIGCHLD and ends the appropriate PIDs if set */
+static int elogind_sigchld_handler(
+                sd_event_source* s,
+                const struct signalfd_siginfo* si,
+                void* userdata ) {
+        Manager* m = userdata;
+        int r, status;
+
+        r = sd_event_get_state( m->event );
+
+        if ( r != SD_EVENT_FINISHED ) {
+                /* Wait for tool PID if set (might have been spawned alone or from sleep/hibernate forker) */
+                if ( m->tool_fork_pid > 0 ) {
+                        waitpid(m->tool_fork_pid, &status, WNOHANG | WUNTRACED);
+                        log_debug_elogind( "tool_fork PID %d waitpid() set status %d", m->tool_fork_pid, status );
+                        if ( WIFEXITED(status) || WIFSIGNALED(status) )
+                                m->tool_fork_pid = 0;
+                }
+
+                /* The sleep forker PID is always "the outer one", so wait for it second. */
+                if ( m->sleep_fork_pid > 0 ) {
+                        waitpid(m->sleep_fork_pid, &status, WNOHANG | WUNTRACED);
+                        log_debug_elogind( "sleep_fork PID %d waitpid() set status %d", m->sleep_fork_pid, status );
+                        if ( WIFEXITED(status) || WIFSIGNALED(status) )
+                                m->sleep_fork_pid = 0;
+                }
+        }
+
+        return 0;
+}
+
+
 static void remove_pid_file( void ) {
-        if ( access( ELOGIND_PID_FILE, F_OK ) == 0 )
-                unlink_noerrno( ELOGIND_PID_FILE );
+        if ( access( ELOGIND_PID_FILE, F_OK ) == 0 ) {
+                int old_errno = errno;
+                unlink( ELOGIND_PID_FILE );
+                if ( old_errno != errno )
+                        errno = old_errno;
+        }
 }
 
 
@@ -116,7 +152,7 @@ static int elogind_daemonize( void ) {
         log_debug_elogind("Parent SID     : %5d", getsid(getpid_cached()));
 
         r = safe_fork( "elogind-forker",
-                       FORK_LOG|FORK_REOPEN_LOG|FORK_DEATHSIG|FORK_CLOSE_ALL_FDS|FORK_NULL_STDIO|FORK_WAIT,
+                       FORK_LOG|FORK_REOPEN_LOG|FORK_DEATHSIG_SIGTERM|FORK_CLOSE_ALL_FDS|FORK_REARRANGE_STDIO|FORK_WAIT,
                        &child );
 
         if ( r ) {
@@ -161,11 +197,11 @@ static int elogind_daemonize( void ) {
         }
 
         if ( r < 0 )
-                return log_error_errno( r, "%s: Failed to fork daemon: %m", program_invocation_short_name );
+                _exit( log_error_errno( r, "%s: Failed to fork daemon: %m", program_invocation_short_name ) );
 
         if ( r )
                 /* Exit immediately! */
-                return r;
+                _exit( EXIT_SUCCESS );
 
         umask( 0022 );
 
@@ -195,7 +231,7 @@ static pid_t elogind_is_already_running( void ) {
                  * process will be "elogind".
                  * Therefore check comm with startswith().
                  */
-                get_process_comm( pid, &comm );
+                pid_get_comm( pid, &comm );
                 if ( NULL == startswith( strna( comm ), program_invocation_short_name ) )
                         goto we_are_alone;
                 return pid;
@@ -286,7 +322,7 @@ int elogind_setup_cgroups_agent( Manager* m ) {
                 (void) unlink( sa.un.sun_path );
 
                 /* Only allow root to connect to this socket */
-                RUN_WITH_UMASK( 0077 )
+                WITH_UMASK( 0077 )
                 r = bind( fd, &sa.sa, SOCKADDR_UN_LEN( sa.un ) );
                 if ( r < 0 )
                         return log_error_errno( errno, "bind(%s) failed: %m", sa.un.sun_path );
@@ -384,6 +420,8 @@ int elogind_manager_new( Manager* m ) {
         m->pin_cgroupfs_fd  = -1;
         m->test_run_flags   = 0;
         m->do_interrupt     = false;
+        m->sleep_fork_pid   = 0;
+        m->tool_fork_pid    = 0;
 
         /* Init poweroff/suspend interruption */
         m->allow_poweroff_interrupts     = false;
@@ -401,8 +439,7 @@ int elogind_manager_new( Manager* m ) {
         m->allow[SLEEP_HIBERNATE]              = true;
         m->allow[SLEEP_HYBRID_SLEEP]           = true;
         m->allow[SLEEP_SUSPEND_THEN_HIBERNATE] = true;
-        m->hibernate_delay_usec                = 0;
-        m->suspend_estimation_usec             = 0;
+        m->hibernate_delay_usec                 = 0;
 
         /* If elogind should be its own controller, mount its cgroup */
         if ( streq( SYSTEMD_CGROUP_CONTROLLER, "_elogind" ) ) {
@@ -442,5 +479,11 @@ int elogind_manager_startup( Manager* m ) {
                 log_error_errno( r, "Failed to register SIGTERM handler: %m" );
         }
 
+        r = sd_event_add_signal( m->event, NULL, SIGCHLD, elogind_sigchld_handler, m );
+        if ( r < 0 ) {
+                if ( e == 0 ) e = r;
+                log_error_errno( r, "Failed to register SIGCHLD handler: %m" );
+        }
+
         return e;
 }
