@@ -2,14 +2,16 @@
 
 #include <fcntl.h>
 //#include <sys/ioctl.h>
-#include <sys/types.h>
+//#include <sys/types.h>
 //#include <linux/vt.h>
 
 #include "sd-device.h"
 
 #include "alloc-util.h"
-//#include "bus-error.h"
-//#include "bus-util.h"
+#include "battery-util.h"
+#include "bus-error.h"
+#include "bus-locator.h"
+#include "bus-util.h"
 #include "cgroup-util.h"
 #include "conf-parser.h"
 #include "device-util.h"
@@ -21,7 +23,7 @@
 //#include "parse-util.h"
 #include "path-util.h"
 #include "process-util.h"
-#include "stdio-util.h"
+// #include "stdio-util.h"
 #include "strv.h"
 //#include "terminal-util.h"
 #include "udev-util.h"
@@ -86,26 +88,19 @@ int manager_parse_config_file(Manager *m) {
         assert(m);
 
 #if 0 /// elogind parses its own config file
-        return config_parse_many_nulstr(
-                        PKGSYSCONFDIR "/logind.conf",
-                        CONF_PATHS_NULSTR("systemd/logind.conf.d"),
-                        "Login\0",
-                        config_item_perf_lookup, logind_gperf_lookup,
-                        CONFIG_PARSE_WARN, m,
-                        NULL);
+        return config_parse_config_file("logind.conf", "Login\0",
+                                        config_item_perf_lookup, logind_gperf_lookup,
+                                        CONFIG_PARSE_WARN, m);
 #else // 0
         const char* logind_conf = getenv("ELOGIND_CONF_FILE");
 
         if (!logind_conf)
                 logind_conf = PKGSYSCONFDIR "/logind.conf";
 
-        return config_parse_many_nulstr(
-                        logind_conf,
-                        CONF_PATHS_NULSTR("elogind/logind.conf.d"),
-                        "Login\0Sleep\0",
+        return config_parse_config_file(
+                        logind_conf, "Login\0",
                         config_item_perf_lookup, logind_gperf_lookup,
-                        CONFIG_PARSE_WARN, m,
-                        NULL);
+                        CONFIG_PARSE_WARN, m);
 #endif // 0
 }
 
@@ -348,15 +343,11 @@ int manager_process_button_device(Manager *m, sd_device *d) {
                 return r;
 
         if (device_for_action(d, SD_DEVICE_REMOVE) ||
-            sd_device_has_current_tag(d, "power-switch") <= 0) {
+            sd_device_has_current_tag(d, "power-switch") <= 0)
 
-                b = hashmap_get(m->buttons, sysname);
-                if (!b)
-                        return 0;
+                button_free(hashmap_get(m->buttons, sysname));
 
-                button_free(b);
-
-        } else {
+        else {
                 const char *sn;
 
                 r = manager_add_button(m, sysname, &b);
@@ -377,7 +368,7 @@ int manager_process_button_device(Manager *m, sd_device *d) {
         return 0;
 }
 
-int manager_get_session_by_pid(Manager *m, pid_t pid, Session **ret) {
+int manager_get_session_by_pidref(Manager *m, const PidRef *pid, Session **ret) {
 #if 0 /// elogind does not support systemd units, but its own session system
         _cleanup_free_ char *unit = NULL;
 #else // 0
@@ -388,18 +379,24 @@ int manager_get_session_by_pid(Manager *m, pid_t pid, Session **ret) {
 
         assert(m);
 
-        if (!pid_is_valid(pid))
+        if (!pidref_is_set(pid))
                 return -EINVAL;
 
-        s = hashmap_get(m->sessions_by_leader, PID_TO_PTR(pid));
-        if (!s) {
+        s = hashmap_get(m->sessions_by_leader, pid);
+        if (s) {
+                r = pidref_verify(pid);
+                if (r < 0)
+                        return r;
+        } else {
 #if 0 /// elogind does not support systemd units, but its own session system
-                r = cg_pid_get_unit(pid, &unit);
-                if (r >= 0)
-                        s = hashmap_get(m->session_units, unit);
+                r = cg_pidref_get_unit(pid, &unit);
+                if (r < 0)
+                        return r;
+
+                s = hashmap_get(m->session_units, unit);
 #else // 0
-                log_debug_elogind("Searching session for PID %d", pid);
-                r = cg_pid_get_session(pid, &session_name);
+                log_debug_elogind("Searching session for PID %d", pid->pid);
+                r = cg_pid_get_session(pid->pid, &session_name);
 
                 if (r >= 0)
                         s = hashmap_get(m->sessions, session_name);
@@ -434,7 +431,7 @@ int manager_get_user_by_pid(Manager *m, pid_t pid, User **ret) {
 
 #if 1 /// elogind also has its own session system
         if ((r < 0) || (NULL == u)) {
-                r = manager_get_session_by_pid(m, pid, &s);
+                r = manager_get_session_by_pidref(m, &PIDREF_MAKE_FROM_PID(pid), &s);
                 // If a session was found, ignore it if it is already closing.
                 if ((r > 0) && (SESSION_CLOSING != session_get_state(s)))
                         u = s->user;
@@ -542,7 +539,7 @@ int config_parse_n_autovts(
 static int vt_is_busy(unsigned vtnr) {
         struct vt_stat vt_stat;
         int r;
-        _cleanup_close_ int fd = -1;
+        _cleanup_close_ int fd = -EBADF;
 
         assert(vtnr >= 1);
 
@@ -592,15 +589,7 @@ int manager_spawn_autovt(Manager *m, unsigned vtnr) {
         }
 
         xsprintf(name, "autovt@tty%u.service", vtnr);
-        r = sd_bus_call_method(
-                        m->bus,
-                        "org.freedesktop.systemd1",
-                        "/org/freedesktop/systemd1",
-                        "org.freedesktop.systemd1.Manager",
-                        "StartUnit",
-                        &error,
-                        NULL,
-                        "ss", name, "fail");
+        r = bus_call_method(m->bus, bus_systemd_mgr, "StartUnit", &error, NULL, "ss", name, "fail");
         if (r < 0)
                 return log_error_errno(r, "Failed to start %s: %s", name, bus_error_message(&error, r));
 
@@ -630,7 +619,6 @@ static bool manager_is_docked(Manager *m) {
 
 static int manager_count_external_displays(Manager *m) {
         _cleanup_(sd_device_enumerator_unrefp) sd_device_enumerator *e = NULL;
-        sd_device *d;
         int r, n = 0;
 
         r = sd_device_enumerator_new(&e);
@@ -800,8 +788,7 @@ int manager_read_utmp(Manager *m) {
                 if (isempty(t))
                         continue;
 
-                s = hashmap_get(m->sessions_by_leader, PID_TO_PTR(u->ut_pid));
-                if (!s)
+                if (manager_get_session_by_pidref(m, &PIDREF_MAKE_FROM_PID(u->ut_pid), &s) <= 0)
                         continue;
 
                 if (s->tty_validity == TTY_FROM_UTMP && !streq_ptr(s->tty, t)) {
@@ -894,13 +881,14 @@ int manager_read_efi_boot_loader_entries(Manager *m) {
                 return 0;
 
         r = efi_loader_get_entries(&m->efi_boot_loader_entries);
-        if (r == -ENOENT || ERRNO_IS_NOT_SUPPORTED(r)) {
-                log_debug_errno(r, "Boot loader reported no entries.");
-                m->efi_boot_loader_entries_set = true;
-                return 0;
-        }
-        if (r < 0)
+        if (r < 0) {
+                if (r == -ENOENT || ERRNO_IS_NOT_SUPPORTED(r)) {
+                        log_debug_errno(r, "Boot loader reported no entries.");
+                        m->efi_boot_loader_entries_set = true;
+                        return 0;
+                }
                 return log_error_errno(r, "Failed to determine entries reported by boot loader: %m");
+        }
 
         m->efi_boot_loader_entries_set = true;
         return 1;
