@@ -13,7 +13,7 @@
 
 #include "alloc-util.h"
 #include "audit-util.h"
-#include "bus-error.h"
+// #include "bus-error.h"
 #include "bus-util.h"
 #include "devnum-util.h"
 #include "env-file.h"
@@ -21,6 +21,7 @@
 #include "fd-util.h"
 #include "fileio.h"
 #include "format-util.h"
+#include "fs-util.h"
 #include "io-util.h"
 #include "logind-dbus.h"
 #include "logind-seat-dbus.h"
@@ -33,12 +34,11 @@
 #include "process-util.h"
 #include "serialize.h"
 #include "string-table.h"
-#include "strv.h"
+// #include "strv.h"
 #include "terminal-util.h"
 #include "tmpfile-util.h"
-#include "uid-alloc-range.h"
+// #include "uid-alloc-range.h"
 #include "user-util.h"
-//#include "util.h"
 /// Additional includes needed by elogind
 #include "cgroup-setup.h"
 #include "extract-word.h"
@@ -65,10 +65,11 @@ int session_new(Session **ret, Manager *m, const char *id) {
 
         *s = (Session) {
                 .manager = m,
-                .fifo_fd = -1,
-                .vtfd = -1,
+                .fifo_fd = -EBADF,
+                .vtfd = -EBADF,
                 .audit_id = AUDIT_SESSION_INVALID,
                 .tty_validity = _TTY_VALIDITY_INVALID,
+                .leader = PIDREF_NULL,
         };
 
         s->state_file = path_join("/run/systemd/sessions", id);
@@ -89,6 +90,17 @@ int session_new(Session **ret, Manager *m, const char *id) {
         return 0;
 }
 
+static void session_reset_leader(Session *s) {
+        assert(s);
+
+        if (!pidref_is_set(&s->leader))
+                return;
+
+        (void) hashmap_remove_value(s->manager->sessions_by_leader, &s->leader, s);
+
+        return pidref_done(&s->leader);
+}
+
 Session* session_free(Session *s) {
         SessionDevice *sd;
 
@@ -134,13 +146,13 @@ Session* session_free(Session *s) {
         }
 #endif // 0
 
-        if (pid_is_valid(s->leader))
-                (void) hashmap_remove_value(s->manager->sessions_by_leader, PID_TO_PTR(s->leader), s);
 
 #if 0 /// elogind does not support systemd scope_jobs
         free(s->scope_job);
 #endif // 0
 
+        session_reset_leader(s);
+
         sd_bus_message_unref(s->create_message);
 
         free(s->tty);
@@ -175,26 +187,26 @@ void session_set_user(Session *s, User *u) {
         user_update_last_session_timer(u);
 }
 
-int session_set_leader(Session *s, pid_t pid) {
+int session_set_leader_consume(Session *s, PidRef _leader) {
+        _cleanup_(pidref_done) PidRef pidref = _leader;
         int r;
 
         assert(s);
+        assert(pidref_is_set(&pidref));
 
-        if (!pid_is_valid(pid))
-                return -EINVAL;
-
-        if (s->leader == pid)
+        if (pidref_equal(&s->leader, &pidref))
                 return 0;
 
-        r = hashmap_put(s->manager->sessions_by_leader, PID_TO_PTR(pid), s);
+        session_reset_leader(s);
+
+        s->leader = TAKE_PIDREF(pidref);
+
+        r = hashmap_ensure_put(&s->manager->sessions_by_leader, &pidref_hash_ops, &s->leader, s);
         if (r < 0)
                 return r;
+        assert(r > 0);
 
-        if (pid_is_valid(s->leader))
-                (void) hashmap_remove_value(s->manager->sessions_by_leader, PID_TO_PTR(s->leader), s);
-
-        s->leader = pid;
-        (void) audit_session_from_pid(pid, &s->audit_id);
+        (void) audit_session_from_pid(s->leader.pid, &s->audit_id);
 
         return 1;
 }
@@ -205,13 +217,13 @@ static void session_save_devices(Session *s, FILE *f) {
         if (!hashmap_isempty(s->devices)) {
                 fprintf(f, "DEVICES=");
                 HASHMAP_FOREACH(sd, s->devices)
-                        fprintf(f, "%u:%u ", major(sd->dev), minor(sd->dev));
+                        fprintf(f, DEVNUM_FORMAT_STR " ", DEVNUM_FORMAT_VAL(sd->dev));
                 fprintf(f, "\n");
         }
 }
 
 int session_save(Session *s) {
-        _cleanup_free_ char *temp_path = NULL;
+        _cleanup_(unlink_and_freep) char *temp_path = NULL;
         _cleanup_fclose_ FILE *f = NULL;
         int r;
 
@@ -333,8 +345,8 @@ int session_save(Session *s) {
         if (!s->vtnr)
                 fprintf(f, "POSITION=%u\n", s->position);
 
-        if (pid_is_valid(s->leader))
-                fprintf(f, "LEADER="PID_FMT"\n", s->leader);
+        if (pidref_is_set(&s->leader))
+                fprintf(f, "LEADER="PID_FMT"\n", s->leader.pid);
 
         if (audit_session_is_valid(s->audit_id))
                 fprintf(f, "AUDIT=%"PRIu32"\n", s->audit_id);
@@ -360,14 +372,12 @@ int session_save(Session *s) {
                 goto fail;
         }
 
+        temp_path = mfree(temp_path);
         return 0;
 
 fail:
         (void) unlink(s->state_file);
 
-        if (temp_path)
-                (void) unlink(temp_path);
-
         return log_error_errno(r, "Failed to save session data %s: %m", s->state_file);
 }
 
@@ -528,13 +538,13 @@ int session_load(Session *s) {
         }
 
         if (leader) {
-                pid_t pid;
+                _cleanup_(pidref_done) PidRef p = PIDREF_NULL;
 
-                r = parse_pid(leader, &pid);
+                r = pidref_set_pidstr(&p, leader);
                 if (r < 0)
                         log_debug_errno(r, "Failed to parse leader PID of session: %s", leader);
                 else {
-                        r = session_set_leader(s, pid);
+                        r = session_set_leader_consume(s, TAKE_PIDREF(p));
                         if (r < 0)
                                 log_warning_errno(r, "Failed to set session leader PID, ignoring: %m");
                 }
@@ -709,7 +719,7 @@ static int session_start_scope(Session *s, sd_bus_message *properties, sd_bus_er
                 r = manager_start_scope(
                                 s->manager,
                                 scope,
-                                s->leader,
+                                &s->leader,
                                 s->user->slice,
                                 description,
                                 /* These two have StopWhenUnneeded= set, hence add a dep towards them */
@@ -737,16 +747,16 @@ static int session_start_cgroup(Session *s) {
 
         assert(s);
         assert(s->user);
-        assert(s->leader > 0);
+        assert(s->leader.pid > 0);
 
         /* First, create our own group */
         r = cg_create(SYSTEMD_CGROUP_CONTROLLER, s->id);
         if (r < 0)
                 return log_error_errno(r, "Failed to create cgroup %s: %m", s->id);
 
-        r = cg_attach(SYSTEMD_CGROUP_CONTROLLER, s->id, s->leader);
+        r = cg_attach(SYSTEMD_CGROUP_CONTROLLER, s->id, s->leader.pid);
         if (r < 0)
-                log_warning_errno(r, "Failed to attach PID %d to cgroup %s: %m", s->leader, s->id);
+                log_warning_errno(r, "Failed to attach PID %d to cgroup %s: %m", s->leader.pid, s->id);
 
         return 0;
 }
@@ -838,11 +848,11 @@ int session_start(Session *s, sd_bus_message *properties, sd_bus_error *error) {
                    "MESSAGE_ID=" SD_MESSAGE_SESSION_START_STR,
                    "SESSION_ID=%s", s->id,
                    "USER_ID=%s", s->user->user_record->user_name,
-                   "LEADER="PID_FMT, s->leader,
+                   "LEADER="PID_FMT, s->leader.pid,
                    LOG_MESSAGE("New session %s of user %s.", s->id, s->user->user_record->user_name));
 
         if (!dual_timestamp_is_set(&s->timestamp))
-                dual_timestamp_get(&s->timestamp);
+                dual_timestamp_now(&s->timestamp);
 
         if (s->seat)
                 seat_read_active_vt(s->seat);
@@ -909,7 +919,7 @@ static int session_stop_scope(Session *s, bool force) {
                 log_struct(s->class == SESSION_BACKGROUND ? LOG_DEBUG : LOG_INFO,
                            "SESSION_ID=%s", s->id,
                            "USER_ID=%s", s->user->user_record->user_name,
-                           "LEADER="PID_FMT, s->leader,
+                           "LEADER="PID_FMT, s->leader.pid,
                            LOG_MESSAGE("Session %s logged out. Waiting for processes to exit.", s->id));
         }
 
@@ -953,6 +963,7 @@ int session_stop(Session *s, bool force) {
                 return 0;
 
         s->timer_event_source = sd_event_source_unref(s->timer_event_source);
+        s->leader_pidfd_event_source = sd_event_source_unref(s->leader_pidfd_event_source);
 
         if (s->seat)
                 seat_evict_position(s->seat, s);
@@ -991,7 +1002,7 @@ int session_finalize(Session *s) {
                            "MESSAGE_ID=" SD_MESSAGE_SESSION_STOP_STR,
                            "SESSION_ID=%s", s->id,
                            "USER_ID=%s", s->user->user_record->user_name,
-                           "LEADER="PID_FMT, s->leader,
+                           "LEADER="PID_FMT, s->leader.pid,
                            LOG_MESSAGE("Removed session %s.", s->id));
 #if 1 /// let elogind at least put out a debug message if s was not started
         else
@@ -999,6 +1010,7 @@ int session_finalize(Session *s) {
 #endif // 1
 
         s->timer_event_source = sd_event_source_unref(s->timer_event_source);
+        s->leader_pidfd_event_source = sd_event_source_unref(s->leader_pidfd_event_source);
 
         if (s->seat)
                 seat_evict_position(s->seat, s);
@@ -1023,6 +1035,8 @@ int session_finalize(Session *s) {
                 seat_save(s->seat);
         }
 
+        session_reset_leader(s);
+
         user_save(s->user);
         user_send_changed(s->user, "Display", NULL);
 
@@ -1130,8 +1144,8 @@ int session_get_idle_hint(Session *s, dual_timestamp *t) {
 
         /* For sessions with a leader but no explicitly configured tty, let's check the controlling tty of
          * the leader */
-        if (pid_is_valid(s->leader)) {
-                r = get_process_ctty_atime(s->leader, &atime);
+        if (pidref_is_set(&s->leader)) {
+                r = get_process_ctty_atime(s->leader.pid, &atime);
                 if (r >= 0)
                         goto found_atime;
         }
@@ -1167,7 +1181,7 @@ int session_set_idle_hint(Session *s, bool b) {
                 return 0;
 
         s->idle_hint = b;
-        dual_timestamp_get(&s->idle_hint_timestamp);
+        dual_timestamp_now(&s->idle_hint_timestamp);
 
         session_send_changed(s, "IdleHint", "IdleSinceHint", "IdleSinceHintMonotonic", NULL);
 
@@ -1226,6 +1240,23 @@ int session_set_display(Session *s, const char *display) {
         return 1;
 }
 
+int session_set_tty(Session *s, const char *tty) {
+        int r;
+
+        assert(s);
+        assert(tty);
+
+        r = free_and_strdup(&s->tty, tty);
+        if (r <= 0)  /* 0 means the strings were equal */
+                return r;
+
+        session_save(s);
+
+        session_send_changed(s, "TTY", NULL);
+
+        return 1;
+}
+
 static int session_dispatch_fifo(sd_event_source *es, int fd, uint32_t revents, void *userdata) {
         Session *s = ASSERT_PTR(userdata);
 
@@ -1311,10 +1342,38 @@ static void session_remove_fifo(Session *s) {
         }
 }
 
+static int session_dispatch_leader_pidfd(sd_event_source *es, int fd, uint32_t revents, void *userdata) {
+        Session *s = ASSERT_PTR(userdata);
+
+        assert(s->leader.fd == fd);
+        session_stop(s, /* force= */ false);
+
+        return 1;
+}
+
+int session_watch_pidfd(Session *s) {
+        int r;
+
+        assert(s);
+
+        if (s->leader.fd < 0)
+                return 0;
+
+        r = sd_event_add_io(s->manager->event, &s->leader_pidfd_event_source, s->leader.fd, EPOLLIN, session_dispatch_leader_pidfd, s);
+        if (r < 0)
+                return r;
+
+        r = sd_event_source_set_priority(s->leader_pidfd_event_source, SD_EVENT_PRIORITY_IMPORTANT);
+        if (r < 0)
+                return r;
+
+        (void) sd_event_source_set_description(s->leader_pidfd_event_source, "session-pidfd");
+
+        return 0;
+}
+
 bool session_may_gc(Session *s, bool drop_not_started) {
-#if 0 /// elogind supports neither scopes nor jobs
         int r;
-#endif // 0
 
         assert(s);
 
@@ -1332,6 +1391,12 @@ bool session_may_gc(Session *s, bool drop_not_started) {
         if (!s->user)
                 return true;
 
+        r = pidref_is_alive(&s->leader);
+        if (r < 0)
+                log_debug_errno(r, "Unable to determine if leader PID " PID_FMT " is still alive, assuming not.", s->leader.pid);
+        if (r > 0)
+                return false;
+
         if (s->fifo_fd >= 0) {
                 if (pipe_eof(s->fifo_fd) <= 0)
                         return false;
@@ -1383,7 +1448,7 @@ SessionState session_get_state(Session *s) {
                 return SESSION_CLOSING;
 
 #if 0 /// elogind does not support systemd scope_jobs
-        if (s->scope_job || s->fifo_fd < 0)
+        if (s->scope_job || (!pidref_is_set(&s->leader) && s->fifo_fd < 0))
 #else // 0
         if (s->fifo_fd < 0)
 #endif // 0
@@ -1405,36 +1470,42 @@ int session_kill(Session *s, KillWho who, int signo) {
         return manager_kill_unit(s->manager, s->scope, who, signo, NULL);
 #else // 0
         if (who == KILL_LEADER) {
-                if (s->leader <= 0)
+                if (s->leader.pid <= 0)
                         return -ESRCH;
 
                 /* FIXME: verify that leader is in cgroup?  */
 
-                if (kill(s->leader, signo) < 0) {
-                        return log_error_errno(errno, "Failed to kill process leader %d for session %s: %m", s->leader, s->id);
+                if (kill(s->leader.pid, signo) < 0) {
+                        return log_error_errno(errno, "Failed to kill process leader %d for session %s: %m", s->leader.pid, s->id);
                 }
                 return 0;
         } else
-                return cg_kill_recursive (SYSTEMD_CGROUP_CONTROLLER, s->id, signo,
+                // const char *path, int sig, CGroupFlags flags, Set *s, cg_kill_log_func_t kill_log, void *userdata
+                return cg_kill_recursive (SYSTEMD_CGROUP_CONTROLLER, signo,
                                           CGROUP_IGNORE_SELF | CGROUP_REMOVE,
                                           NULL, NULL, NULL);
 #endif // 0
 }
 
-static int session_open_vt(Session *s) {
+static int session_open_vt(Session *s, bool reopen) {
+        _cleanup_close_ int fd = -EBADF;
         char path[sizeof("/dev/tty") + DECIMAL_STR_MAX(s->vtnr)];
 
+        assert(s);
+
         if (s->vtnr < 1)
                 return -ENODEV;
 
-        if (s->vtfd >= 0)
+        if (!reopen && s->vtfd >= 0)
                 return s->vtfd;
 
         sprintf(path, "/dev/tty%u", s->vtnr);
-        s->vtfd = open_terminal(path, O_RDWR | O_CLOEXEC | O_NONBLOCK | O_NOCTTY);
-        if (s->vtfd < 0)
-                return log_error_errno(s->vtfd, "cannot open VT %s of session %s: %m", path, s->id);
 
+        fd = open_terminal(path, O_RDWR | O_CLOEXEC | O_NONBLOCK | O_NOCTTY);
+        if (fd < 0)
+                return log_error_errno(fd, "Cannot open VT %s of session %s: %m", path, s->id);
+
+        close_and_replace(s->vtfd, fd);
         return s->vtfd;
 }
 
@@ -1442,10 +1513,12 @@ static int session_prepare_vt(Session *s) {
         int vt, r;
         struct vt_mode mode = {};
 
+        assert(s);
+
         if (s->vtnr < 1)
                 return 0;
 
-        vt = session_open_vt(s);
+        vt = session_open_vt(s, /* reopen = */ false);
         if (vt < 0)
                 return vt;
 
@@ -1497,25 +1570,24 @@ error:
 static void session_restore_vt(Session *s) {
         int r;
 
+        assert(s);
+
+        if (s->vtfd < 0)
+                return;
+
         r = vt_restore(s->vtfd);
         if (r == -EIO) {
-                int vt, old_fd;
-
                 /* It might happen if the controlling process exited before or while we were
                  * restoring the VT as it would leave the old file-descriptor in a hung-up
                  * state. In this case let's retry with a fresh handle to the virtual terminal. */
 
                 /* We do a little dance to avoid having the terminal be available
                  * for reuse before we've cleaned it up. */
-                old_fd = TAKE_FD(s->vtfd);
 
-                vt = session_open_vt(s);
-                safe_close(old_fd);
-
-                if (vt >= 0)
-                        r = vt_restore(vt);
+                int fd = session_open_vt(s, /* reopen = */ true);
+                if (fd >= 0)
+                        r = vt_restore(fd);
         }
-
         if (r < 0)
                 log_warning_errno(r, "Failed to restore VT, ignoring: %m");
 
@@ -1542,23 +1614,13 @@ void session_leave_vt(Session *s) {
                 return;
 
         session_device_pause_all(s);
-        r = vt_release(s->vtfd, false);
+        r = vt_release(s->vtfd, /* restore = */ false);
         if (r == -EIO) {
-                int vt, old_fd;
+                /* Handle the same VT hung-up case as in session_restore_vt */
 
-                /* It might happen if the controlling process exited before or while we were
-                 * restoring the VT as it would leave the old file-descriptor in a hung-up
-                 * state. In this case let's retry with a fresh handle to the virtual terminal. */
-
-                /* We do a little dance to avoid having the terminal be available
-                 * for reuse before we've cleaned it up. */
-                old_fd = TAKE_FD(s->vtfd);
-
-                vt = session_open_vt(s);
-                safe_close(old_fd);
-
-                if (vt >= 0)
-                        r = vt_release(vt, false);
+                int fd = session_open_vt(s, /* reopen = */ true);
+                if (fd >= 0)
+                        r = vt_release(fd, /* restore = */ false);
         }
         if (r < 0)
                 log_debug_errno(r, "Cannot release VT of session %s: %m", s->id);
@@ -1572,6 +1634,8 @@ static void session_release_controller(Session *s, bool notify) {
         _unused_ _cleanup_free_ char *name = NULL;
         SessionDevice *sd;
 
+        assert(s);
+
         if (!s->controller)
                 return;
 
